/*
 *  Copyright (C) 2007, 2008  A.N. Yzelman
 *
 *  Last modified at 25th of February, 2008, by A.N. Yzelman
 */

#include "Matrix2HilbertCoordinates.h"

/**
 *  Convenience method for creating 2D std::vectors
 *
 *  @param x First dimension value
 *  @param y Second dimension value
 *  @param where Where to store the requested std::vector
 */
void Matrix2HilbertCoordinates::makeVec( const double &x, const double &y, std::vector< double >* where ) {
	(*where)[ 0 ] = x;
	(*where)[ 1 ] = y;
}

/**
 *  Function returning the euclidian distance between two std::vectors.
 *
 *  @param vx Pointer to the first std::vector
 *  @param vy Pointer to the second std::vector
 *  @param dim The number of dimensions
 *  @return The euclidian distance; sqrt(<vx,vy>)
 */
double Matrix2HilbertCoordinates::dist( const std::vector< double > *vx, const std::vector< double > *vy, const unsigned int &dim ) {

	double ret = ( (*vx)[ 0 ] - (*vy)[ 0 ] ); ret *= ret;
	for ( unsigned int i=1; i<dim; i++ ) {
		const double cur = ( (*vx)[ i ] - (*vy)[ i ] );
		ret += cur*cur;
	}
	
	//return sqrt( ret );
	return ret; //speed you fool!
}

/**
 *  Determines the closest std::vector from a set to a given std::vector,
 *  in terms of euclidian distance.
 *
 *  @param corners The set of coordinates we wish to select the closest from
 *  @param coor The coordinate we wish to check against
 *  @param dim The number of dimensions
 *  @return The index of the closest coordinate in coors
 */
unsigned long int Matrix2HilbertCoordinates::closestTo( const std::vector< double > *coor, const std::vector< std::vector< double > > *corners ) {
	const unsigned long int dim = 2;

	if (corners->size() == 0 )
		std::cerr << "Empty corners std::vector passed to Matrix2HilbertCoordinates::closestTo()!" << std::endl;
	
	unsigned long int ret = 0;
	double min = dist( coor, &( corners->at( 0 ) ), dim );
	
	
	for ( unsigned int i=1 ; i<corners->size(); i++ ) {
		const double curcost = dist( coor, & ( corners->at( i ) ), dim );
		if ( curcost < min ) {
			min = curcost;
			ret = i;
		}
	}
	
	return ret;
}

/**
 *  Calculates the midpoint of two std::vectors. Contains an O(1) inefficiency in the form of
 *  asking the std::vector for the current dimension.
 *
 *  @param x A pointer to the first std::vector
 *  @param y A pointer to the second std::vector
 *  @return The std::vector containing the mid point
 */
void Matrix2HilbertCoordinates::mid( const std::vector< double > *x, const std::vector< double > *y, std::vector< double > *to ) {
	
	double* toArr = & ( (*to)[ 0 ] );
	for( unsigned int i=0; i<x->size(); i++ )
		toArr[ i ] = ( (*x)[ i ] + (*y)[ i ] ) / 2.0;

}

/**
 *  Function which refines a given section of a rectangle.
 *
 *	@param corners A std::vector containing the 4 cornerpoints of a rectangle. At the end of the function call, this contains the 4 corners of the refined rectangle at the given cell.
 *  @param cell A number [0,1,..,4] indicating which cell we are refining (refer to the paper for more information).
 */
void Matrix2HilbertCoordinates::refineCoors2D( std::vector< std::vector< double > > *corners, const unsigned int cell ) {

	std::vector< double > a( 2 );
	std::vector< double > b( 2 );
	std::vector< double > c( 2 );
	std::vector< double > d( 2 );

	std::vector< double > *in = &( ( *corners )[ 0 ] );

	//assume order of the std::vector to be alphabetical
	switch( cell ) {
		case 1:
			a = in[ 0 ];
			mid( &in[ 0 ], &in[ 3 ], &b );
			mid( &in[ 0 ], &in[ 2 ], &c );
			mid( &in[ 0 ], &in[ 1 ], &d );
			break;
		case 2:
			mid( &in[ 0 ], &in[ 1 ], &a );
			b = in[ 1 ];
			mid( &in[ 1 ], &in[ 2 ], &c );
			mid( &in[ 1 ], &in[ 3 ], &d );
			break;
		case 3:
			mid( &in[ 0 ], &in[ 2 ], &a );
			mid( &in[ 1 ], &in[ 2 ], &b );
			c = in[ 2 ];
			mid( &in[ 2 ], &in[ 3 ], &d );
			break;
		case 4:
			mid( &in[ 3 ], &in[ 2 ], &a );
			mid( &in[ 1 ], &in[ 3 ], &b );
			mid( &in[ 3 ], &in[ 0 ], &c );
			d = in[ 3 ];
			break;
		default:
			std::cerr << "Invalid cell no. given in refineHilbert2D!" << std::endl;
			std::exit( 1 );
	}
	
	(*corners)[ 0 ] = a;
	(*corners)[ 1 ] = b;
	(*corners)[ 2 ] = c;
	(*corners)[ 3 ] = d;

}

/**
 *  Maps a 2D matrix entry position to a fractional number in base 4 stored in the vector 'digits';
 *  this function will increase the number of digits by one.
 *
 *  @param i The row-coordinate of the matrix entry.
 *  @param j The column-coordinate of the matrix entry.
 *  @param m The number of matrix rows.
 *  @param n The number of matrix columns.
 *  @param digits The current digits of the Hilbert coordinate.
 */
void Matrix2HilbertCoordinates::getDigit( 	const unsigned long int &i, const unsigned long int &j, 
						const unsigned long int &m, const unsigned long int &n,
						std::vector< unsigned long int > &digits ) {
	std::vector< std::vector< double > > Box;
	std::vector< double > coordinates( 2 );

	coordinates[ 0 ] = static_cast< double >( i ) / static_cast< double >( m - 1 );
	coordinates[ 1 ] = static_cast< double >( j ) / static_cast< double >( n - 1 );

	//note: remember that the order matters!
	std::vector< double > a( 2 ); makeVec( 0, 0, &a); Box.push_back( a );
	std::vector< double > b( 2 ); makeVec( 0, 1, &b); Box.push_back( b );
	std::vector< double > c( 2 ); makeVec( 1, 1, &c); Box.push_back( c );
	std::vector< double > d( 2 ); makeVec( 1, 0, &d); Box.push_back( d );

	unsigned long int closest;

	for( unsigned long int i=0; i<digits.size(); i++ ) {
		closest = digits[ i ];
		refineCoors2D( &Box, closest+1 );
	}
	
	//double curDist;
	closest = closestTo( &coordinates, &Box );
	digits.push_back( closest );
	
}

