/*
 *  Copyright (C) 2007  A.N. Yzelman
 *
 *  Last modified at 15th of February, 2008, by A.N. Yzelman
 */


#ifndef _H_Matrix2HilbertCoordinates
#define _H_Matrix2HilbertCoordinates

#include <cmath>
#include <vector>
#include <iostream>
#include <cstdlib>

/**
 *  Class which maps coordinates to 1D Hilbert Coordinates.
 */
class Matrix2HilbertCoordinates {

   private:

	/** Base constructor. Does nothing. */
	Matrix2HilbertCoordinates() {}
   
	/** Convenience 2D vector creation function. */
	static void makeVec( const double &x, const double &y, std::vector< double > *where );

	/** 
	 *  Calculates the midpoint of two vectors.
	 *  
	 *  @param x The first vector.
	 *  @param y The second vector.
	 *  @param to Pointer to where the mid-point vector should be stored in.
	 */
	static void mid( const std::vector< double > *x, const std::vector< double > *y, std::vector< double > *to );

	/**
	 *  Calculates the norm between two vectors of a given dimension.
	 *  @param x Pointer to the first vector.
	 *  @param y Pointer to the second vector.
	 *  @param dim Reference to the dimension of the vectors.
	 *  @return The normed distance.
	 */
	static double dist( const std::vector< double > *x, const std::vector< double > *y, const unsigned int &dim );

	/**
	 *  Determines which of the corners a given vector is closest to.
	 *  @param coor Pointer to the vector to determine the closeness of.
	 *  @param corners Coordinates of the individual corners.
	 *  @return The index number of the closest corner.
	 */
	static unsigned long int closestTo( const std::vector< double > *coor, const std::vector< std::vector< double > > *corners );

	/**
	 *  Refines the block we are working on by zooming in on a given cell.
	 *  @param corners Coordinates of the current corners (these will be updated).
	 *  @param cell The cell we will zoom in on.
	 */
	static void refineCoors2D( std::vector< std::vector< double > > *corners, const unsigned int cell );

	/** Translates a base 4 digit to base 10
	 *  @param A vector of fractional digits (.1135--) in base 4.
	 *  @return A base-10 double representation of the given digits.
	 */
	static double base4_to_base10( const std::vector< unsigned int > *digits ) {
		return base_n_to_base10< 4 >( digits );
	}

	/** Translate digits from any base to base 10.
	 *  (used in higher dimensions; >2)
	 *  @param base the base from which to covert.
	 *  @param digits pointer to a vector containing the fractional digits.
	 *  @return A base-10 representation of the given number.
	 */
	template< const unsigned long int base >
	static double base_n_to_base10( const std::vector< unsigned int > *digits ) {
		const unsigned long int numdigits = digits->size();
		double ret = 0;
		unsigned long int deler = 1;
	
		for( unsigned long int i = 0; i < numdigits; i++ ) {
			const unsigned long int noemer = (*digits)[ i ];
			deler *= ( unsigned long int )base;
			ret += ( ( double ) noemer ) / ( ( double ) deler );
		}
	
		return ret;
	}

   protected:

   public:

	/** Extends a Hilbert-coordinate by adding one digit.
	 *  @param i An index value in one dimension, 0<=i<m
	 *  @param j An index value in the other dimension, 0<=j<n
	 *  @param m Max index value in one dimension
	 *  @param n Max index value in the other dimension
	 *  @param digits reference on which vector to add the next Hilbert coordinate digit.
	 */
	static void getDigit( 		const unsigned long int &i, const unsigned long int &j,
					const unsigned long int &m, const unsigned long int &n,
					std::vector< unsigned long int > &digits
				 );

};

#endif

